#include <fstream>
#include <vector>
#include <valarray>
#include <math.h>
#include <stdlib.h>
#include <string>
#include <iostream>
#include <iomanip>
#include <complex>
#include <thread>
#include <random>

#ifdef DOUBLE
typedef double data_type;
#else
typedef float data_type;
#endif

#include "constants.h"
#include "fft.h"
#include "interpolation.h"
#include "thread_functions.h"
#include "prime_numbers.h"

using namespace std;

/*
 * For a thorough documentation see "Grid.cpp".
 */
class Grid {

 public:

  Grid(int seed);
  ~Grid();
  
  void print_shear_catalog(int n, data_type sigma_eps, int number_of_grid, string file_name);
  void print_clustered_galaxies(int n, data_type bias, data_type bias_quadratic, int number_of_grid, string file_name);
  void print_random_points(int number_of_randoms, string file_name);
  void print_shear_grid(string file_name, int number_of_grid);
  void print_delta_grid(string file_name, int number_of_grid);
  void show_status();
  
 private:
      
   int number_of_steps;
   int number_of_threads;
   int number_of_delta_fields;
   int number_of_shear_fields;
   int tolerance;
   int parts_ra;
   int parts_de;
   int seed_value;
   
   vector<int> log_normal_delta;
   vector<int> log_normal_shear;
   vector<int> reduced_shears;
   vector<data_type> kappa_zeros;
   
   double variance_of_Gaussian_delta_field;
   vector<double> variances_of_Gaussian_delta_fields;
   
   data_type theta_max;
   data_type l_min;
   data_type l_max;
   
   valarray<data_type> ln_l;
   valarray<data_type> thetas;
   valarray<data_type> positions;
   vector<valarray<valarray<complex<data_type> > > > random_fields;
   
   // The ij-th entry is the power spectrum between field i and field j.
   vector<vector<valarray<data_type> > > power_spectra;
   
   // In order to enable parallel processing, the shear grid
   // has to be split into several sub grids:
   vector<valarray<data_type> > sub_thetas_ra;
   vector<valarray<data_type> > sub_thetas_de;
   vector<vector<data_type> > sub_positions;
   vector<vector<data_type> > sub_random_numbers;
   vector<vector<valarray<data_type> > > sub_galaxy_data;
   vector<valarray<valarray<complex<data_type> > > > sub_grids;
   
   void configure(string config_name);
   void configure_delta_fields(string config_name);
   void configure_shear_fields(string config_name);
   
   void read_power_spectra(string power_spectrum_name);
   void log_normal_tranform_to_power_spectra();
   void check_l_range();
   void set_number_of_threads();
   void set_grid();
   void set_fourier_and_spatial_grids(vector<valarray<valarray<complex<data_type> > > > *temporary_fields);
   void set_sub_grids(int n, int number_of_grid, bool intrinsic_noise);
   
};
