
/* _____ Content Grid_configure.cpp
 * 
 * ..... 1. configure
 * ..... 2. configure_delta_fields
 * ..... 3. configure_shear_fields
 * 
 */

/*******************************************************************************************************************************************************
 * 1. configure
 * Description:
 * - Reads general configuration file.
 * Arguments:
 * - config_name: name of general configuration file
 * 
*******************************************************************************************************************************************************/

void Grid::configure(string config_name){

  int powers_of_2;
  data_type dummy_variable;
  
  // reading the configuration file
  fstream str;
  str.open(config_name.c_str());
  str >> dummy_variable; str >> this->theta_max;
  str >> dummy_variable; str >> this->l_max;
  str >> dummy_variable; str >> powers_of_2;
  str >> dummy_variable; str >> this->number_of_delta_fields;
  str >> dummy_variable; str >> this->number_of_shear_fields;
  str >> dummy_variable; str >> this->number_of_threads;
  str.close();

  this->number_of_steps  = int(pow(2.0, powers_of_2)+0.1);
  this->reduced_shears = vector<int>(this->number_of_shear_fields, 0);
  this->log_normal_delta = vector<int>(this->number_of_delta_fields, 0);
  this->log_normal_shear = vector<int>(this->number_of_shear_fields, 0);
  this->kappa_zeros = vector<data_type>(this->number_of_shear_fields, 0.0);

}


/*******************************************************************************************************************************************************
 * 2. configure_delta_fields
 * Description:
 * - Reads configuration file for the delta fields.
 * Arguments:
 * - config_name: name of configuration file
 * 
*******************************************************************************************************************************************************/

void Grid::configure_delta_fields(string config_name){

  data_type dummy_variable;
  
  // reading the configuration file
  fstream str;
  str.open(config_name.c_str());
  string first_line;
  getline(str, first_line);
  
  for(int i = 0; i < this->number_of_delta_fields; i++){
    str >> dummy_variable;
    str >> this->log_normal_delta[i];
  }
  
  str.close();

}


/*******************************************************************************************************************************************************
 * 3. configure_shear_fields
 * Description:
 * - Reads configuration file for the shear fields.
 * Arguments:
 * - config_name: name of configuration file
 * 
*******************************************************************************************************************************************************/

void Grid::configure_shear_fields(string config_name){

  string rest_of_line;
  data_type dummy_variable;
  
  // reading the configuration file
  fstream str;
  str.open(config_name.c_str());
  string first_line;
  getline(str, first_line);
  
  for(int i = 0; i < this->number_of_shear_fields; i++){
    str >> dummy_variable;
    str >> this->reduced_shears[i];
    str >> this->log_normal_shear[i];
    if(this->log_normal_shear[i] == 1)
      str >> this->kappa_zeros[i];
    else
      getline(str, rest_of_line);
      
  }
  
  
  str.close();

}
