
/* _____ Content Grid_generate_random_fields.cpp
 * 
 * ..... 1. set_grid
 * ..... 2. set_fourier_grid
 * 
 */

/*******************************************************************************************************************************************************
 * 1. set_grid
 * Description:
 * - Creates a temporary array and passes it over to set_fourier_grid.
 * Arguments:
 * -
 * 
*******************************************************************************************************************************************************/


void Grid::set_grid(){
  int n = this->number_of_steps;
  vector<valarray<valarray<complex<data_type> > > > temporary_fields;
  this->set_fourier_and_spatial_grids(&temporary_fields);
  
  cout << "random fields set.\n\n";
}


/*******************************************************************************************************************************************************
 * 2. set_fourier_and_spatial_grids
 * Description:
 * - Creates all required random fields.
 * Arguments:
 * - temporary_fields: pointer to space for all the random fields.
 * 
*******************************************************************************************************************************************************/

void Grid::set_fourier_and_spatial_grids(vector<valarray<valarray<complex<data_type> > > > *temporary_fields){
  
  /*
   * Initializing all variables
   */
  int n = this->number_of_steps;
  int index;
  int seed = this->seed_value;
  int n_of_threads = this->number_of_threads;
  int n_fields = this->number_of_delta_fields + this->number_of_shear_fields;
   
  int new_n = 0;
  int offset_n;
  data_type dtheta = constants::pi/this->l_max;
  
  // From the transformed spatial grid, only a sub grid will be used:
  
  dtheta /= constants::arcmin;

  for(int j = 0; j < n, data_type(j)*dtheta <= this->theta_max; j++)
    new_n += 1;
      
  // also cut out a tolerance region around the new grid
  // for better interpolation on the grid edges:
  if(new_n + 2*this->tolerance > n){
    this->tolerance = (n - new_n)/2;
    new_n = n;
  }
  else
    new_n += 2*this->tolerance;
  offset_n = (n-new_n)/2;
  this->thetas.resize(new_n, 0.0);
  for(int i = 0; i < new_n; i++)
    this->thetas[i] = data_type(i-this->tolerance)*dtheta;
  
  while((n/2)%n_of_threads!=0)
    n_of_threads++;
  
  data_type k, k1, k2;
  data_type phi;
  data_type cos_phi, sin_phi;
  data_type s;
  data_type u1, u2;
  data_type r, im;
  data_type dk = 2*this->l_max/data_type(this->number_of_steps);
  
  mt19937_64 gen(seed);
  
  vector<valarray<double> > sigma_gauss(n_fields, valarray<double>(0.0,n/2+1));
  vector<valarray<valarray<data_type> > > random_numbers(n_fields, valarray<valarray<data_type> >(valarray<data_type>(0.0, 2*n), n/2+1));

  /*
   * Generating the random numbers that will be handed
   * over to the different threads.
   */
  data_type normalization = data_type(gen.max());
  for(int i = 0; i < n_fields; i++){
    thread random_thread = thread(generate_random_numbers, n, &random_numbers[i], &gen, normalization, i, n_fields);
    (*temporary_fields).push_back(valarray<valarray<complex<data_type> > >(valarray<complex<data_type> >(0.0, n), n));
    random_thread.join();
  }
  cout << "\rGenerating random numbers:      100%     \n";
  /*
   * Drawing the modes in one half of the Fourier plane
   * from a Gaussian probability distribution.
   */
  for(int i = 0; i < n/2; i+=n_of_threads){
    cout << "\rGenerating Gaussian fields:     "  << data_type(1000*i/(n-2))/10 << "%";
    cout.flush();
    
    vector<thread> th;
    for(int j = 0; j < n_of_threads; j++){
      th.push_back(thread(compute_k_modes, dk, data_type(i+j)*dk,
			  this->l_min, this->l_max, this->ln_l,
			  this->power_spectra,
			  &(*temporary_fields),
			  i+j,
			  &random_numbers,
			  &sigma_gauss));
    }
    for(int j = 0; j < n_of_threads; j++)
      th[j].join();
    
    th.clear();
    
  }
  // Compute last mode array:
  compute_k_modes(dk, data_type(n/2)*dk,
		 this->l_min, this->l_max, this->ln_l,
		 this->power_spectra,
		 &(*temporary_fields),
		 n/2,
		 &random_numbers,
		 &sigma_gauss);
  random_numbers.clear();
  /*
   * Choose modes in second half of the Fourier plane such
   * that kappa is a real function in position space.
   */  
  
  
  // respect symmetries of the Fourier transform of a real function.
  for(int i = 0; i < n_fields; i++){

    for(int j = n/2; j < n; j+=n_of_threads){
      cout << "\rGenerating Gaussian fields:     " << data_type(1000*(j-n/2)/(n_fields*n)+ 500*i/n_fields)/10 + 50 << "%";
      cout.flush();
  
      vector<thread> th;
      for(int k = 0; k < n_of_threads; k++)
        th.push_back(thread(respect_symmetries, &(*temporary_fields)[i][j+k], &(*temporary_fields)[i][n-(j+k)]));
      for(int k = 0; k < n_of_threads; k++)
        th[k].join();
    }
    for(int j = 1; j <= n/2; j++){
      (*temporary_fields)[i][0][j] = conj((*temporary_fields)[i][0][n-j]);
    }
    (*temporary_fields)[i][0][0] = 0.0;
    (*temporary_fields)[i][n-n/2][n-n/2] = (*temporary_fields)[i][n-n/2][n-n/2].real()*sqrt(2.0);
    (*temporary_fields)[i][n-n/2][0] = (*temporary_fields)[i][n-n/2][0].real()*sqrt(2.0);
    (*temporary_fields)[i][0][n-n/2] = (*temporary_fields)[i][0][n-n/2].real()*sqrt(2.0);
  }
  
  cout << "\rGenerating Gaussian fields:     100%    \n";

  /*
   * generate delta fields
   */
  this->variances_of_Gaussian_delta_fields = vector<double>(0,0.0);

  for(int i = 0; i < this->number_of_delta_fields; i++){
    
    this->variance_of_Gaussian_delta_field = 0.0;
    for(int j = 0; j < n/2; j++){
      this->variance_of_Gaussian_delta_field += 2.0*sigma_gauss[i][j];
    }
    this->variance_of_Gaussian_delta_field += sigma_gauss[i][n/2];
    sigma_gauss[i].resize(0);
    this->variances_of_Gaussian_delta_fields.push_back(this->variance_of_Gaussian_delta_field);
    
    if(this->log_normal_delta[i] == 1){
           fft_2D(&(*temporary_fields)[i], n, n, this->number_of_threads, 1, i);
	    double mu = -this->variance_of_Gaussian_delta_field/2.0;
	    
	    for(int j = 0; j < n; j+= n_of_threads){
	      cout << "\rGenerating delta field " << i+1 << ":       "  << data_type(200*j/n)/10 + 80 << "%";
	      cout.flush();
	      
	      vector<thread> th;
	    
	      for(int k = 0; k < n_of_threads; k++)
	        th.push_back(thread(transform_from_Gaussian_to_lognorm, &(*temporary_fields)[i][j+k], mu, 1.0));
	        
	      for(int k = 0; k < n_of_threads; k++)
	        th[k].join();
	    }
	}
	else{
	  fft_2D(&(*temporary_fields)[i], n, n, this->number_of_threads, 3, i);
	}
     
    this->random_fields.push_back(valarray<valarray<complex<data_type> > >(valarray<complex<data_type> >(0.0, new_n), new_n));
    for(int k = 0; k < new_n; k++){
      this->random_fields[i][k] = (*temporary_fields)[i][k+offset_n][slice(offset_n,new_n,1)];
    }
    /*
    double test = 0.0;
    for(int k = 0; k < new_n; k++){
      double testtest = 0.0;
      for(int kk = 0; kk < new_n; kk++){
        testtest += double(this->random_fields[i][k][kk].real());
      }
      testtest /= double(new_n);
      test += testtest;
    }
    test /= double(new_n);
    cout << setprecision(10);
    cout << "mean delta field " << i+1 << ": " << test << "\n";
    cout << setprecision(2);
    */
    cout << "\rGenerating delta field " << i+1 << ":       100%        \n";
    if(sqrt(this->variance_of_Gaussian_delta_field) >= 0.3 && this->log_normal_delta[i] != 1)
      cerr << "Warning: variance of Gaussian delta field is close to 1 (sqrt(var) >= 0.3).\n"; 


  }

  /*
   * generate kappa fields
   */

  for(int i = this->number_of_delta_fields; i < n_fields; i++){
    
    valarray<valarray<complex<data_type> > > kappa_field_spatial;
    
    if(this->log_normal_shear[i-this->number_of_delta_fields] == 1){
        fft_2D(&(*temporary_fields)[i], n, n, this->number_of_threads, 2, i-this->number_of_delta_fields);
	    this->variance_of_Gaussian_delta_field = 0.0;
	    for(int j = 0; j < n/2; j++){
	      this->variance_of_Gaussian_delta_field += 2*sigma_gauss[i][j];
	    }
	    this->variance_of_Gaussian_delta_field += sigma_gauss[i][n/2];
	    data_type k0 = this->kappa_zeros[i-this->number_of_delta_fields];
	    double mu = log(double(k0))-this->variance_of_Gaussian_delta_field/2.0;
	    
	    for(int j = 0; j < n; j+= n_of_threads){
	      cout << "\rGenerating kappa field " << i+1-this->number_of_delta_fields << ":       "  << data_type(200*j/n)/10 + 40 << "%";
	      cout.flush();
	      
	      vector<thread> th;
	    
	      for(int k = 0; k < n_of_threads; k++)
	        th.push_back(thread(transform_from_Gaussian_to_lognorm, &(*temporary_fields)[i][j+k], mu, k0));
	        
	      for(int k = 0; k < n_of_threads; k++)
	        th[k].join();
	    }
	    
	    if(this->reduced_shears[i-this->number_of_delta_fields] == 1){
	      kappa_field_spatial = valarray<valarray<complex<data_type> > >(valarray<complex<data_type> >(0.0, new_n), new_n);
	      for(int k = 0; k < new_n; k++){
		kappa_field_spatial[k] = (*temporary_fields)[i][k+offset_n][slice(offset_n,new_n,1)];
	      }
	    }
	    
	    fft_2D_inverse(&(*temporary_fields)[i], n, n, this->number_of_threads, 1, i-this->number_of_delta_fields); 
	    cout << "\rGenerating kappa field " << i+1-this->number_of_delta_fields << ":       100%        \n";
	}
    sigma_gauss[i].resize(0);
    
     
    /*
     * Now change the Fourier transform of kappa into the Fourier transform of gamma
     * by multiplying the appropriate phase factor.
     */
    if(this->log_normal_shear[i-this->number_of_delta_fields] == 0 && this->reduced_shears[i-this->number_of_delta_fields] == 1){
      valarray<valarray<complex<data_type> > > kappa_field_fourier;
      kappa_field_fourier = (*temporary_fields)[i];
      fft_2D(&kappa_field_fourier, n, n, this->number_of_threads, 2, i-this->number_of_delta_fields);
      kappa_field_spatial = valarray<valarray<complex<data_type> > >(valarray<complex<data_type> >(0.0, new_n), new_n);
      for(int k = 0; k < new_n; k++){
	kappa_field_spatial[k] = kappa_field_fourier[k+offset_n][slice(offset_n,new_n,1)];
      }
    }
    
    for(int j = 0; j < n; j+= n_of_threads){
      cout << "\rGenerating shear field " << i+1-this->number_of_delta_fields << ":       "  << data_type(400*j/n)/10 << "%";
      cout.flush();
       
      vector<thread> th;
      
      for(int k = 0; k < n_of_threads; k++)
        th.push_back(thread(apply_phase_factor, &(*temporary_fields)[i][j+k], dk, this->l_max, j+k));
        
      for(int k = 0; k < n_of_threads; k++)
        th[k].join();
    }
    fft_2D(&(*temporary_fields)[i], n, n, this->number_of_threads, 4, i-this->number_of_delta_fields);
    this->random_fields.push_back(valarray<valarray<complex<data_type> > >(valarray<complex<data_type> >(0.0, new_n), new_n));
    for(int k = 0; k < new_n; k++){
      this->random_fields[i][k] = (*temporary_fields)[i][k+offset_n][slice(offset_n,new_n,1)];
    }
    cout << "\rGenerating shear field " << i+1-this->number_of_delta_fields << ":       100%        \n";

    if(this->reduced_shears[i-this->number_of_delta_fields] == 1){
      for(int j = 0; j < new_n; j+= n_of_threads){
        cout << "\rReducing shear field " << i+1-this->number_of_delta_fields << ":         "  << data_type(400*j/n)/10 << "%";
        cout.flush();

	vector<thread> th;

	for(int k = 0; k < n_of_threads; k++)
	  if(j+k < new_n)
	    th.push_back(thread(reduce_shear_grid, &this->random_fields[i][j+k], &kappa_field_spatial[j+k]));
	
	for(int k = 0; k < n_of_threads; k++){
	  cout.flush();
	  if(j+k < new_n){
	    th[k].join();
	  }
	}
      }
      cout << "\rReducing shear grid " << i+1-this->number_of_delta_fields << ":          100%       \n";

    }
    
  }
  
}

/*******************************************************************************************************************************************************
 * 2.1 set_sub_grids
 * Description:
 * - Divides the shear grid into several sub grids in order to use multithreading in the method
 *   print_random_shear_catalog. Also, the theta ranges that correspond to the different sub grids
 *   are store and the position of the galaxies are drawn and distributed to the sub grids.
 * - Also stores the random numbers that will in print_random_shear_catalog be used to draw the
 *   intrinsic ellipticity noise of each galaxy, if sigma_eps > eps.
 *   (If sigma_eps < eps, the intrinsic noise will simply be ignored.)
 *
 * Arguments:
 * - n: number of galaxies that shall be placed on the grid
 * - seed: seed of the RNG that is used to draw the galaxy positions.
 * 
 * 
*******************************************************************************************************************************************************/

void Grid::set_sub_grids(int n, int number_of_grid, bool intrinsic_noise){
  cout << "Setting sub grids";
  
  int index;
  int length;
  int seed = this->seed_value;
  this->seed_value = next_prime(this->seed_value);
  this->seed_value = next_prime(this->seed_value);
  int length_ra = 0;
  int length_de = 0;
  this->parts_ra = 0;
  this->parts_de = 0;
  int thread_number = this->number_of_threads;
  int check = 1;
  int subgrid_overlap = this->tolerance;
  if(this->tolerance < 8)
    subgrid_overlap = 8;
  
  data_type dtheta = this->thetas[1] - this->thetas[0];
  data_type dra;
  data_type dde;
  data_type ra;
  data_type de;
  mt19937_64 gen(seed);
  data_type normalization = data_type(gen.max())/this->theta_max;
  
    // finding out in how many parts the field has to be split up in ra and de:
  
  while(thread_number > 1 && thread_number%2 == 0){
    
    thread_number /= 2;
    if(check == 1)
      this->parts_de += 1;
    else
      this->parts_ra += 1;
    
    check *= -1;
    
  }
  
  this->parts_de = pow(2, this->parts_de)*thread_number;
  this->parts_ra = pow(2, this->parts_ra);
  // side length of the sub regions:
  dra = this->theta_max/data_type(this->parts_ra);
  dde = this->theta_max/data_type(this->parts_de);
  
  // integer length of the sub grids that will be cut out:
  length_ra = int(dra/dtheta);
  length_de = int(dde/dtheta);
  
  if(this->parts_de*this->parts_ra != this->number_of_threads){
    cerr << "Splitting into subgrids was not successful!\nSee method \"set_sub_grids()\" for details.\n";
    exit(EXIT_FAILURE);
  }  
  
  
  // initializing the sub theta array, which will be used for interpolation
  // within the sub grids:
  cout.flush();
  this->sub_thetas_ra = vector<valarray<data_type> > (this->parts_ra, valarray<data_type>(0.0, length_ra+2*subgrid_overlap));
  this->sub_thetas_de = vector<valarray<data_type> > (this->parts_de, valarray<data_type>(0.0, length_de+2*subgrid_overlap));
  
  for(int i = 0; i < this->parts_ra; i++){
    int ra_start = i*length_ra + this->tolerance - subgrid_overlap;
    if(ra_start < 0) ra_start = 0;
    while(ra_start + length_ra+2*subgrid_overlap > this->thetas.size()) ra_start--;
    this->sub_thetas_ra[i] = this->thetas[slice(ra_start, length_ra+2*subgrid_overlap, 1)];
  }
  
  for(int j = 0; j < this->parts_de; j++){
    int de_start = j*length_de + this->tolerance - subgrid_overlap;
    if(de_start < 0) de_start = 0;
    while(de_start + length_de+2*subgrid_overlap > this->thetas.size()) de_start--;
    this->sub_thetas_de[j] = this->thetas[slice(de_start, length_de+2*subgrid_overlap, 1)];
  }
  
  // cutting out the subgrids:
  
  this->sub_grids = vector<valarray<valarray<complex<data_type> > > >(this->number_of_threads, valarray<valarray<complex<data_type> > >(valarray<complex<data_type> >(0.0, length_de+2*subgrid_overlap),length_ra+2*subgrid_overlap));
  
  for(int i = 0; i < this->parts_ra; i++){
    int ra_start = i*length_ra + this->tolerance - subgrid_overlap;
    if(ra_start < 0) ra_start = 0;
    while(ra_start + length_ra+2*subgrid_overlap > this->thetas.size()) ra_start--;
    
    for(int j = 0; j < this->parts_de; j++){
    int de_start = j*length_de + this->tolerance - subgrid_overlap;
    if(de_start < 0) de_start = 0;
    while(de_start+length_de+2*subgrid_overlap > this->thetas.size()) de_start--;
      for(int k = 0; k < this->sub_grids[i+j*this->parts_ra].size(); k++){
	this->sub_grids[i+j*this->parts_ra][k] = this->random_fields[number_of_grid][k+ra_start][slice(de_start,length_de+2*subgrid_overlap,1)];
      }
    }
  }
  
  // initializing the positions of the galaxies and
  // distributing them to the ra and de sub grids:
  
  this->sub_positions = vector<vector<data_type> >(this->number_of_threads, vector<data_type>(0, 0.0));
  this->sub_random_numbers = vector<vector<data_type> >(this->number_of_threads, vector<data_type>(0, 0));
  
  for(int i = 0; i < n; i++){
    ra = data_type(gen())/normalization;
    de = data_type(gen())/normalization;
    int patch_ra = int(ra/dra);
    int patch_de = int(de/dde);      
    index = patch_ra + patch_de*this->parts_ra;
    if(index >= this->number_of_threads || patch_ra >= this->parts_ra || patch_de >= this->parts_de)
      i -= 1;
    else{
      this->sub_positions[index].push_back(ra);
      this->sub_positions[index].push_back(de);
    }
  }

  if(intrinsic_noise){
    for(int i = 0; i < this->number_of_threads; i++){
    length = this->sub_positions[i].size();
    this->sub_random_numbers[i].resize(length);
      for(int j = 0; j < length; j++)
        this->sub_random_numbers[i][j] = data_type(gen());
    }
  }
  
  
  // preparing the array that will store galaxy data:
  length = 2*int(data_type(n)/data_type(this->number_of_threads));
  this->sub_galaxy_data = 
  vector<vector<valarray<data_type> > >(this->number_of_threads, vector<valarray<data_type> >(0, valarray<data_type>(2)));
  
  for(int i = 0; i < this->number_of_threads; i++)
    this->sub_galaxy_data[i].resize(this->sub_positions[i].size()/2, valarray<data_type>(0.0,2));
  
  cout << " done.\n";
}


