#include "Hankel_Transformer.h"
#include "hankel.h"

using namespace std;

Hankel_Transformer::Hankel_Transformer(int log_x, int log_y, valarray<double> x_in, valarray<double> y_in){
  
  this->set_x_and_y(log_x, log_y, x_in, y_in);
  
}



Hankel_Transformer::~Hankel_Transformer(){
  
  
}


void Hankel_Transformer::set_x_and_y(int log_x, int log_y, valarray<double> x_in, valarray<double> y_in){
  
  int length = x_in.size();

  double dummy_variable;
  
  this->x = valarray<double>(0.0, length);
  this->y = valarray<double>(0.0, length);
  
  for(int i = 0; i < length; i++){
    if(log_x == 1)
      this->x[i] = exp(x_in[i]);
    else
      this->x[i] = x_in[i];

    if(log_y == 1)
      this->y[i] = exp(y_in[i]);
    else
      this->y[i] = y_in[i];
  }
  
}



void Hankel_Transformer::lognorm_to_Gaussian_P(double k0, valarray<double>* P){
  
  valarray<double> theta;
  valarray<double> xi;
  valarray<double> ell;
  valarray<double> l_save = this->x;
  valarray<double> p_save = this->y;
  
  this->P_to_xi(0, &theta, &xi);
  this->lognorm_to_Gaussian_xi(k0, &xi);
  
  this->x = theta;
  this->y = xi;
  
  this->xi_to_P(0, &ell, P);
  
  this->x = l_save;
  this->y = p_save;

}


void Hankel_Transformer::lognorm_to_Gaussian_xi(double k0, valarray<double>* xi){
  
  int n = (*xi).size();
  double one_over_k0_sq = 1.0/(k0*k0);
  
  for(int i = 0; i < n; i++){
    if((*xi)[i] <= -1.0)
      cerr << "AHHHHH!!!!!  " << (*xi)[i] << '\n';
    (*xi)[i] = log((*xi)[i]*one_over_k0_sq+1.0);
  }
  
}
void Hankel_Transformer::P_to_xi(int nu, valarray<double>* x_out, valarray<double>* y_out){
  
  hankel::Hankel_transform(&this->x, &this->y, x_out, y_out, nu, 0);
  
}
void Hankel_Transformer::xi_to_P(int nu, valarray<double>* x_out, valarray<double>* y_out){
  
  hankel::Hankel_transform(&this->x, &this->y, x_out, y_out, nu, 1);
  
}