using namespace std;

/*******************************************************************************************************************************************************
 * fft
 * 
 * Description:
 * - Fast Fourier transform of the array (*f).
 * 
 * Arguments:
 * - f: pointer on the array to be transformed
 * - e: stored exponential factor appearing in the Fourier transform.
 * 
 * Comments:
 *   - The exponential factor "e" is provided from outside such that
 *     the exponential function does not have to be computed in every
 *     call of fft.
 * 
*******************************************************************************************************************************************************/

void fft_double(valarray<complex<double> > *f, complex<double> e){
 
 
  int n2 = (*f).size();
  
  if(n2 > 1){
    
    int n = n2/2;
    
    complex<double> one = 1.0;
    complex<double> two = 1.0;
    complex<double> three = 1.0;
    
    valarray<complex<double> > f1(n);
    valarray<complex<double> > f2(n);
    
    f1 = (*f)[slice(0,n,2)];
    f2 = (*f)[slice(1,n,2)];
    
    fft_double(&f1, e*e);
    fft_double(&f2, e*e);
    
    for(int i = 0; i < n; i++){
      two = f1[i];
      three = one*f2[i];
      (*f)[i] =  two + three; 
      (*f)[i+n] = two - three;
      one *= e;
    }
    
    
  }
  
}
