#include <complex>
#include <math.h>
#include <iostream>
#include <valarray>
#include <gsl/gsl_sf_gamma.h>
#include <gsl/gsl_complex.h>
#include "fft_double.h"

using namespace std;


namespace hankel{

void hankel_kernel_Fourier(valarray<double> *s, valarray<complex<double> > *H, int nu){
  
  int n = (*s).size();
  
  gsl_sf_result ln;
  gsl_sf_result arg;
  double ln_2 = log(2.0);
  
  complex<double> z0;

  for(int i = 0; i < n; i++){
    

    gsl_sf_lngamma_complex_e(0.5*(1.0+double(nu)), 0.5*(*s)[i], &ln, &arg);

    if(arg.err/arg.val >1.0e-8)
      cerr << "HAHAHAHAHAHA\n";
    
    (*H)[i] = exp(complex<double>(0,2*arg.val+ln_2*(*s)[i]));
  }
  
}


void Hankel_transform(valarray<double> *x_input, 
			   valarray<double> *y_input,
			   valarray<double> *x_output, 
			   valarray<double> *y_output,
			   int nu,
			   int switch_to_inverse){

  int n = 2*(*y_input).size();
  valarray<complex<double> > y(0.0, n);
  valarray<double> x(0.0, n);
  for(int i = 0; i < n/2; i++){
    x[i] = (*x_input)[i];
    y[i] = (*y_input)[i];
  }
  
  double x_min = x[0];
  double x_max = x[n/2-1];
  double l_min = log(x_min);
  double l_max = log(x_max);
  double dl = log(x_max/x_min)/double(n/2-1);
  double ds = 2*constants::pi/dl/double(n);
  double prefactor = 1.0;
  complex<double> im(0,2*constants::pi/double(n));
  complex<double> e = exp(im);
  complex<double> e_bar = exp(-im);
  valarray<double> l(0.0, n);
  valarray<double> s(0.0, n/2+1);
  valarray<complex<double> > H(0.0, n/2+1);
  valarray<complex<double> > p_small(0.0, n);
  
  
  for(int i = n/2; i < n; i++){
    x[i] = exp(l_min+double(i)*dl);
  }
  for(int i = 0; i <= n/2; i++){
    s[i] = double(i)*ds;
  }  
  for(int i = 0; i < n; i++){
    p_small[i] = y[n-1-i]*x[n-1-i];
  }
  p_small = p_small.cshift(-n/2);
  fft_double(&p_small, e);
  
  hankel::hankel_kernel_Fourier(&s, &H, nu);
  for(int i = 0; i <= n/2; i++){
    p_small[i] *= H[i]/double(n);
  }
  for(int i = n/2; i < n; i++){
    p_small[i] = conj(p_small[n-i]);
  }
  p_small[0] = complex<double>(p_small[0].real(), 0.0);
  p_small[n/2] = complex<double>(p_small[n/2].real(), 0.0);
  fft_double(&p_small, e_bar);
  p_small = p_small.cshift(n/2);
  
  
  if(switch_to_inverse == 0){
    prefactor = 1.0/(2.0*constants::pi);
  }
  else{
    prefactor = 2.0*constants::pi;
  }
  (*x_output) = valarray<double>(0.0, n/2);
  (*y_output) = valarray<double>(0.0, n/2);
  for(int i = 0; i < n/2; i++){
    (*x_output)[i] = 1.0/x[n/2-1-i];
    (*y_output)[i] = p_small[i+n/2].real()*x[n/2-1-i]*prefactor;
  }
  
}




};