using namespace std;

/*******************************************************************************************************************************************************
 * Neville-Aitken Interpolation
 * 
*******************************************************************************************************************************************************/


int find_index(data_type x0, valarray<data_type> *x){

  int i1 = 0;
  int i2 = x->size();
  int pivot;

  while(i2-i1 > 1){
    pivot = (i2+i1)/2;
    if((*x)[pivot]<=x0)
      i1 = pivot;
    else
      i2 = pivot;
  }
  
  if(i1 == (*x).size()-1)
    i1--;

  return i1;

}


data_type neville_aitken(data_type x0, valarray<data_type> *x, valarray<data_type> *f){
  
  int n = (*x).size();
  
  if(n == 2){
    
    data_type x1 = (*x)[0];
    data_type f1 = (*f)[0];
    
    return f1 + (x0-x1)*((*f)[1]-f1)/((*x)[1]-x1);
  }
  else{
    int n_m1 = n-1;
  
    data_type x1 = (*x)[0];
    data_type x2 = (*x)[n_m1];
  
    valarray<data_type> x_new_1(n_m1);
    valarray<data_type> x_new_2(n_m1);
    valarray<data_type> f_new_1(n_m1);
    valarray<data_type> f_new_2(n_m1);
  
    for(int i = 0; i < n_m1; i++){
       x_new_1[i] = (*x)[i];
       x_new_2[i] = (*x)[i+1];
     
       f_new_1[i] = (*f)[i];
       f_new_2[i] = (*f)[i+1];
    }
  
    return ((x0-x1)*neville_aitken(x0, &x_new_2, &f_new_2)-(x0-x2)*neville_aitken(x0, &x_new_1, &f_new_1))/(x2-x1);
  }
  
  return -1;
  
}

data_type interpolate_neville_aitken(data_type x0, valarray<data_type> *x, valarray<data_type> *f, int order){
  
  int index = find_index(x0, x);
  int n = (*x).size();
  
  valarray<data_type> x_new(order+1);
  valarray<data_type> f_new(order+1);
  
  
  index -= (order+1)/2;
  if(index < 0)
    index = 0;
  if(index + order > n - 1)
    index = n - order - 1;
  
  for(int i = 0; i <= order; i++){
    x_new[i] = (*x)[index+i];
    f_new[i] = (*f)[index+i];
  }
  
  return neville_aitken(x0, &x_new, &f_new);
  
}


complex<data_type> neville_aitken_complex(data_type x0, valarray<data_type> *x, valarray<complex<data_type> > *f){
  
  int n = (*x).size();
  
  if(n == 2){
    
    data_type x1 = (*x)[0];
    complex<data_type> f1 = (*f)[0];
    
    return f1 + (x0-x1)*((*f)[1]-f1)/((*x)[1]-x1);
  }
  else{
    int n_m1 = n-1;
  
    data_type x1 = (*x)[0];
    data_type x2 = (*x)[n_m1];
  
    valarray<data_type> x_new_1(n_m1);
    valarray<data_type> x_new_2(n_m1);
    valarray<complex<data_type> > f_new_1(n_m1);
    valarray<complex<data_type> > f_new_2(n_m1);
  
    for(int i = 0; i < n_m1; i++){
       x_new_1[i] = (*x)[i];
       x_new_2[i] = (*x)[i+1];
     
       f_new_1[i] = (*f)[i];
       f_new_2[i] = (*f)[i+1];
    }
  
    return ((x0-x1)*neville_aitken_complex(x0, &x_new_2, &f_new_2)-(x0-x2)*
		    neville_aitken_complex(x0, &x_new_1, &f_new_1))/(x2-x1);
  }
  
  return -1.0;
  
}

complex<data_type> interpolate_neville_aitken_complex(data_type x0, valarray<data_type> *x, valarray<complex<data_type> > f, int order){
  
  int index = find_index(x0, x);
  int n = (*x).size();
  
  valarray<data_type> x_new(order+1);
  valarray<complex<data_type> > f_new(order+1);
  
  
  index -= (order+1)/2;
  if(index < 0)
    index = 0;
  if(index + order > n - 1)
    index = n - order - 1;
  
  for(int i = 0; i <= order; i++){
    x_new[i] = (*x)[index+i];
    f_new[i] = f[index+i];
  }
  
  return neville_aitken_complex(x0, &x_new, &f_new);
  
}


complex<data_type> interpolate_complex_grid(data_type t0, data_type x0, valarray<data_type> * t, valarray<data_type> * x,
				       valarray<valarray<complex<data_type> > > * f_of_t_x, int order_t, int order_x){

  int index = find_index(t0, t);
  int n_x = (*x).size();
  int n_t = (*t).size();

  valarray<data_type> t_new(order_t+1);
  valarray<complex<data_type> > f_new(order_t+1);
  
  index -= (order_t+1)/2;
  if(index < 0)
    index = 0;
  if(index + order_t > n_t - 1)
    index = n_t - order_t - 1;
  
  for(int i = 0; i <= order_t; i++){
    t_new[i] = (*t)[i+index];
    f_new[i] = interpolate_neville_aitken_complex(x0, x, (*f_of_t_x)[i+index], order_x);
  }
  
  return neville_aitken_complex(t0, &t_new, &f_new);
  
  
}


complex<data_type> interpolate_real_grid(data_type t0, data_type x0, valarray<data_type> * t, valarray<data_type> * x,
				       valarray<valarray<data_type> > * f_of_t_x, int order_t, int order_x){

  int index = find_index(t0, t);
  int n_x = (*x).size();
  int n_t = (*t).size();

  valarray<data_type> t_new(order_t+1);
  valarray<data_type> f_new(order_t+1);
  
  index -= (order_t+1)/2;
  if(index < 0)
    index = 0;
  if(index + order_t > n_t - 1)
    index = n_t - order_t - 1;
  
  for(int i = 0; i <= order_t; i++){
    t_new[i] = (*t)[i+index];
    f_new[i] = interpolate_neville_aitken(x0, x, &(*f_of_t_x)[i+index], order_x);
  }
  
  return neville_aitken(t0, &t_new, &f_new);
  
  
}


